"""
Exports VS15_WIDE_TO_NARROW table keyed by supporting unicode version level.

This code generated by wcwidth/bin/update-tables.py on 2024-02-14 19:59:22 UTC.
"""

VS15_WIDE_TO_NARROW = {
    "9.0.0": (
        # Source: 9.0.0
        # Date: 2023-02-01, 02:22:54 GMT
        #
        (
            0x0231A,
            0x0231B,
        ),  # Watch                   ..Hourglass
        (
            0x023E9,
            0x023EC,
        ),  # Black Right-pointing Dou..Black Down-pointing Doub
        (
            0x023F0,
            0x023F0,
        ),  # Alarm Clock
        (
            0x023F3,
            0x023F3,
        ),  # Hourglass With Flowing Sand
        (
            0x025FD,
            0x025FE,
        ),  # White Medium Small Squar..Black Medium Small Squar
        (
            0x02614,
            0x02615,
        ),  # Umbrella With Rain Drops..Hot Beverage
        (
            0x02648,
            0x02653,
        ),  # Aries                   ..Pisces
        (
            0x0267F,
            0x0267F,
        ),  # Wheelchair Symbol
        (
            0x02693,
            0x02693,
        ),  # Anchor
        (
            0x026A1,
            0x026A1,
        ),  # High Voltage Sign
        (
            0x026AA,
            0x026AB,
        ),  # Medium White Circle     ..Medium Black Circle
        (
            0x026BD,
            0x026BE,
        ),  # Soccer Ball             ..Baseball
        (
            0x026C4,
            0x026C5,
        ),  # Snowman Without Snow    ..Sun Behind Cloud
        (
            0x026CE,
            0x026CE,
        ),  # Ophiuchus
        (
            0x026D4,
            0x026D4,
        ),  # No Entry
        (
            0x026EA,
            0x026EA,
        ),  # Church
        (
            0x026F2,
            0x026F3,
        ),  # Fountain                ..Flag In Hole
        (
            0x026F5,
            0x026F5,
        ),  # Sailboat
        (
            0x026FA,
            0x026FA,
        ),  # Tent
        (
            0x026FD,
            0x026FD,
        ),  # Fuel Pump
        (
            0x02705,
            0x02705,
        ),  # White Heavy Check Mark
        (
            0x0270A,
            0x0270B,
        ),  # Raised Fist             ..Raised Hand
        (
            0x02728,
            0x02728,
        ),  # Sparkles
        (
            0x0274C,
            0x0274C,
        ),  # Cross Mark
        (
            0x0274E,
            0x0274E,
        ),  # Negative Squared Cross Mark
        (
            0x02753,
            0x02755,
        ),  # Black Question Mark Orna..White Exclamation Mark O
        (
            0x02757,
            0x02757,
        ),  # Heavy Exclamation Mark Symbol
        (
            0x02795,
            0x02797,
        ),  # Heavy Plus Sign         ..Heavy Division Sign
        (
            0x027B0,
            0x027B0,
        ),  # Curly Loop
        (
            0x027BF,
            0x027BF,
        ),  # Double Curly Loop
        (
            0x02B1B,
            0x02B1C,
        ),  # Black Large Square      ..White Large Square
        (
            0x02B50,
            0x02B50,
        ),  # White Medium Star
        (
            0x02B55,
            0x02B55,
        ),  # Heavy Large Circle
        (
            0x03030,
            0x03030,
        ),  # Wavy Dash
        (
            0x0303D,
            0x0303D,
        ),  # Part Alternation Mark
        (
            0x03297,
            0x03297,
        ),  # Circled Ideograph Congratulation
        (
            0x03299,
            0x03299,
        ),  # Circled Ideograph Secret
        (
            0x1F004,
            0x1F004,
        ),  # Mahjong Tile Red Dragon
        (
            0x1F202,
            0x1F202,
        ),  # Squared Katakana Sa
        (
            0x1F21A,
            0x1F21A,
        ),  # Squared Cjk Unified Ideograph-7121
        (
            0x1F22F,
            0x1F22F,
        ),  # Squared Cjk Unified Ideograph-6307
        (
            0x1F237,
            0x1F237,
        ),  # Squared Cjk Unified Ideograph-6708
        (
            0x1F30D,
            0x1F30F,
        ),  # Earth Globe Europe-afric..Earth Globe Asia-austral
        (
            0x1F315,
            0x1F315,
        ),  # Full Moon Symbol
        (
            0x1F31C,
            0x1F31C,
        ),  # Last Quarter Moon With Face
        (
            0x1F378,
            0x1F378,
        ),  # Cocktail Glass
        (
            0x1F393,
            0x1F393,
        ),  # Graduation Cap
        (
            0x1F3A7,
            0x1F3A7,
        ),  # Headphone
        (
            0x1F3AC,
            0x1F3AE,
        ),  # Clapper Board           ..Video Game
        (
            0x1F3C2,
            0x1F3C2,
        ),  # Snowboarder
        (
            0x1F3C4,
            0x1F3C4,
        ),  # Surfer
        (
            0x1F3C6,
            0x1F3C6,
        ),  # Trophy
        (
            0x1F3CA,
            0x1F3CA,
        ),  # Swimmer
        (
            0x1F3E0,
            0x1F3E0,
        ),  # House Building
        (
            0x1F3ED,
            0x1F3ED,
        ),  # Factory
        (
            0x1F408,
            0x1F408,
        ),  # Cat
        (
            0x1F415,
            0x1F415,
        ),  # Dog
        (
            0x1F41F,
            0x1F41F,
        ),  # Fish
        (
            0x1F426,
            0x1F426,
        ),  # Bird
        (
            0x1F442,
            0x1F442,
        ),  # Ear
        (
            0x1F446,
            0x1F449,
        ),  # White Up Pointing Backha..White Right Pointing Bac
        (
            0x1F44D,
            0x1F44E,
        ),  # Thumbs Up Sign          ..Thumbs Down Sign
        (
            0x1F453,
            0x1F453,
        ),  # Eyeglasses
        (
            0x1F46A,
            0x1F46A,
        ),  # Family
        (
            0x1F47D,
            0x1F47D,
        ),  # Extraterrestrial Alien
        (
            0x1F4A3,
            0x1F4A3,
        ),  # Bomb
        (
            0x1F4B0,
            0x1F4B0,
        ),  # Money Bag
        (
            0x1F4B3,
            0x1F4B3,
        ),  # Credit Card
        (
            0x1F4BB,
            0x1F4BB,
        ),  # Personal Computer
        (
            0x1F4BF,
            0x1F4BF,
        ),  # Optical Disc
        (
            0x1F4CB,
            0x1F4CB,
        ),  # Clipboard
        (
            0x1F4DA,
            0x1F4DA,
        ),  # Books
        (
            0x1F4DF,
            0x1F4DF,
        ),  # Pager
        (
            0x1F4E4,
            0x1F4E6,
        ),  # Outbox Tray             ..Package
        (
            0x1F4EA,
            0x1F4ED,
        ),  # Closed Mailbox With Lowe..Open Mailbox With Lowere
        (
            0x1F4F7,
            0x1F4F7,
        ),  # Camera
        (
            0x1F4F9,
            0x1F4FB,
        ),  # Video Camera            ..Radio
        (
            0x1F508,
            0x1F508,
        ),  # Speaker
        (
            0x1F50D,
            0x1F50D,
        ),  # Left-pointing Magnifying Glass
        (
            0x1F512,
            0x1F513,
        ),  # Lock                    ..Open Lock
        (
            0x1F550,
            0x1F567,
        ),  # Clock Face One Oclock   ..Clock Face Twelve-thirty
        (
            0x1F610,
            0x1F610,
        ),  # Neutral Face
        (
            0x1F687,
            0x1F687,
        ),  # Metro
        (
            0x1F68D,
            0x1F68D,
        ),  # Oncoming Bus
        (
            0x1F691,
            0x1F691,
        ),  # Ambulance
        (
            0x1F694,
            0x1F694,
        ),  # Oncoming Police Car
        (
            0x1F698,
            0x1F698,
        ),  # Oncoming Automobile
        (
            0x1F6AD,
            0x1F6AD,
        ),  # No Smoking Symbol
        (
            0x1F6B2,
            0x1F6B2,
        ),  # Bicycle
        (
            0x1F6B9,
            0x1F6BA,
        ),  # Mens Symbol             ..Womens Symbol
        (
            0x1F6BC,
            0x1F6BC,
        ),  # Baby Symbol
    ),
}
