"""
Exports VS16_NARROW_TO_WIDE table keyed by supporting unicode version level.

This code generated by wcwidth/bin/update-tables.py on 2025-09-15 16:57:50 UTC.
"""

VS16_NARROW_TO_WIDE = {
    "9.0.0": (
        # Source: 9.0.0
        # Date: 2025-01-30, 21:48:29 GMT
        #
        (
            0x00023,
            0x00023,
        ),  # Number Sign
        (
            0x0002A,
            0x0002A,
        ),  # Asterisk
        (
            0x00030,
            0x00039,
        ),  # Digit Zero              ..Digit Nine
        (
            0x000A9,
            0x000A9,
        ),  # Copyright Sign
        (
            0x000AE,
            0x000AE,
        ),  # Registered Sign
        (
            0x0203C,
            0x0203C,
        ),  # Double Exclamation Mark
        (
            0x02049,
            0x02049,
        ),  # Exclamation Question Mark
        (
            0x02122,
            0x02122,
        ),  # Trade Mark Sign
        (
            0x02139,
            0x02139,
        ),  # Information Source
        (
            0x02194,
            0x02199,
        ),  # Left Right Arrow        ..South West Arrow
        (
            0x021A9,
            0x021AA,
        ),  # Leftwards Arrow With Hoo..Rightwards Arrow With Ho
        (
            0x02328,
            0x02328,
        ),  # Keyboard
        (
            0x023CF,
            0x023CF,
        ),  # Eject Symbol
        (
            0x023ED,
            0x023EF,
        ),  # Black Right-pointing Dou..Black Right-pointing Tri
        (
            0x023F1,
            0x023F2,
        ),  # Stopwatch               ..Timer Clock
        (
            0x023F8,
            0x023FA,
        ),  # Double Vertical Bar     ..Black Circle For Record
        (
            0x024C2,
            0x024C2,
        ),  # Circled Latin Capital Letter M
        (
            0x025AA,
            0x025AB,
        ),  # Black Small Square      ..White Small Square
        (
            0x025B6,
            0x025B6,
        ),  # Black Right-pointing Triangle
        (
            0x025C0,
            0x025C0,
        ),  # Black Left-pointing Triangle
        (
            0x025FB,
            0x025FC,
        ),  # White Medium Square     ..Black Medium Square
        (
            0x02600,
            0x02604,
        ),  # Black Sun With Rays     ..Comet
        (
            0x0260E,
            0x0260E,
        ),  # Black Telephone
        (
            0x02611,
            0x02611,
        ),  # Ballot Box With Check
        (
            0x02618,
            0x02618,
        ),  # Shamrock
        (
            0x0261D,
            0x0261D,
        ),  # White Up Pointing Index
        (
            0x02620,
            0x02620,
        ),  # Skull And Crossbones
        (
            0x02622,
            0x02623,
        ),  # Radioactive Sign        ..Biohazard Sign
        (
            0x02626,
            0x02626,
        ),  # Orthodox Cross
        (
            0x0262A,
            0x0262A,
        ),  # Star And Crescent
        (
            0x0262E,
            0x0262F,
        ),  # Peace Symbol            ..Yin Yang
        (
            0x02638,
            0x0263A,
        ),  # Wheel Of Dharma         ..White Smiling Face
        (
            0x02640,
            0x02640,
        ),  # Female Sign
        (
            0x02642,
            0x02642,
        ),  # Male Sign
        (
            0x0265F,
            0x02660,
        ),  # Black Chess Pawn        ..Black Spade Suit
        (
            0x02663,
            0x02663,
        ),  # Black Club Suit
        (
            0x02665,
            0x02666,
        ),  # Black Heart Suit        ..Black Diamond Suit
        (
            0x02668,
            0x02668,
        ),  # Hot Springs
        (
            0x0267B,
            0x0267B,
        ),  # Black Universal Recycling Symbol
        (
            0x0267E,
            0x0267E,
        ),  # Permanent Paper Sign
        (
            0x02692,
            0x02692,
        ),  # Hammer And Pick
        (
            0x02694,
            0x02697,
        ),  # Crossed Swords          ..Alembic
        (
            0x02699,
            0x02699,
        ),  # Gear
        (
            0x0269B,
            0x0269C,
        ),  # Atom Symbol             ..Fleur-de-lis
        (
            0x026A0,
            0x026A0,
        ),  # Warning Sign
        (
            0x026A7,
            0x026A7,
        ),  # Male With Stroke And Male And Female Sign
        (
            0x026B0,
            0x026B1,
        ),  # Coffin                  ..Funeral Urn
        (
            0x026C8,
            0x026C8,
        ),  # Thunder Cloud And Rain
        (
            0x026CF,
            0x026CF,
        ),  # Pick
        (
            0x026D1,
            0x026D1,
        ),  # Helmet With White Cross
        (
            0x026D3,
            0x026D3,
        ),  # Chains
        (
            0x026E9,
            0x026E9,
        ),  # Shinto Shrine
        (
            0x026F0,
            0x026F1,
        ),  # Mountain                ..Umbrella On Ground
        (
            0x026F4,
            0x026F4,
        ),  # Ferry
        (
            0x026F7,
            0x026F9,
        ),  # Skier                   ..Person With Ball
        (
            0x02702,
            0x02702,
        ),  # Black Scissors
        (
            0x02708,
            0x02709,
        ),  # Airplane                ..Envelope
        (
            0x0270C,
            0x0270D,
        ),  # Victory Hand            ..Writing Hand
        (
            0x0270F,
            0x0270F,
        ),  # Pencil
        (
            0x02712,
            0x02712,
        ),  # Black Nib
        (
            0x02714,
            0x02714,
        ),  # Heavy Check Mark
        (
            0x02716,
            0x02716,
        ),  # Heavy Multiplication X
        (
            0x0271D,
            0x0271D,
        ),  # Latin Cross
        (
            0x02721,
            0x02721,
        ),  # Star Of David
        (
            0x02733,
            0x02734,
        ),  # Eight Spoked Asterisk   ..Eight Pointed Black Star
        (
            0x02744,
            0x02744,
        ),  # Snowflake
        (
            0x02747,
            0x02747,
        ),  # Sparkle
        (
            0x02763,
            0x02764,
        ),  # Heavy Heart Exclamation ..Heavy Black Heart
        (
            0x027A1,
            0x027A1,
        ),  # Black Rightwards Arrow
        (
            0x02934,
            0x02935,
        ),  # Arrow Pointing Rightward..Arrow Pointing Rightward
        (
            0x02B05,
            0x02B07,
        ),  # Leftwards Black Arrow   ..Downwards Black Arrow
        (
            0x1F170,
            0x1F171,
        ),  # Negative Squared Latin C..Negative Squared Latin C
        (
            0x1F17E,
            0x1F17F,
        ),  # Negative Squared Latin C..Negative Squared Latin C
        (
            0x1F321,
            0x1F321,
        ),  # Thermometer
        (
            0x1F324,
            0x1F32C,
        ),  # White Sun With Small Clo..Wind Blowing Face
        (
            0x1F336,
            0x1F336,
        ),  # Hot Pepper
        (
            0x1F37D,
            0x1F37D,
        ),  # Fork And Knife With Plate
        (
            0x1F396,
            0x1F397,
        ),  # Military Medal          ..Reminder Ribbon
        (
            0x1F399,
            0x1F39B,
        ),  # Studio Microphone       ..Control Knobs
        (
            0x1F39E,
            0x1F39F,
        ),  # Film Frames             ..Admission Tickets
        (
            0x1F3CB,
            0x1F3CE,
        ),  # Weight Lifter           ..Racing Car
        (
            0x1F3D4,
            0x1F3DF,
        ),  # Snow Capped Mountain    ..Stadium
        (
            0x1F3F3,
            0x1F3F3,
        ),  # Waving White Flag
        (
            0x1F3F5,
            0x1F3F5,
        ),  # Rosette
        (
            0x1F3F7,
            0x1F3F7,
        ),  # Label
        (
            0x1F43F,
            0x1F43F,
        ),  # Chipmunk
        (
            0x1F441,
            0x1F441,
        ),  # Eye
        (
            0x1F4FD,
            0x1F4FD,
        ),  # Film Projector
        (
            0x1F549,
            0x1F54A,
        ),  # Om Symbol               ..Dove Of Peace
        (
            0x1F56F,
            0x1F570,
        ),  # Candle                  ..Mantelpiece Clock
        (
            0x1F573,
            0x1F579,
        ),  # Hole                    ..Joystick
        (
            0x1F587,
            0x1F587,
        ),  # Linked Paperclips
        (
            0x1F58A,
            0x1F58D,
        ),  # Lower Left Ballpoint Pen..Lower Left Crayon
        (
            0x1F590,
            0x1F590,
        ),  # Raised Hand With Fingers Splayed
        (
            0x1F5A5,
            0x1F5A5,
        ),  # Desktop Computer
        (
            0x1F5A8,
            0x1F5A8,
        ),  # Printer
        (
            0x1F5B1,
            0x1F5B2,
        ),  # Three Button Mouse      ..Trackball
        (
            0x1F5BC,
            0x1F5BC,
        ),  # Frame With Picture
        (
            0x1F5C2,
            0x1F5C4,
        ),  # Card Index Dividers     ..File Cabinet
        (
            0x1F5D1,
            0x1F5D3,
        ),  # Wastebasket             ..Spiral Calendar Pad
        (
            0x1F5DC,
            0x1F5DE,
        ),  # Compression             ..Rolled-up Newspaper
        (
            0x1F5E1,
            0x1F5E1,
        ),  # Dagger Knife
        (
            0x1F5E3,
            0x1F5E3,
        ),  # Speaking Head In Silhouette
        (
            0x1F5E8,
            0x1F5E8,
        ),  # Left Speech Bubble
        (
            0x1F5EF,
            0x1F5EF,
        ),  # Right Anger Bubble
        (
            0x1F5F3,
            0x1F5F3,
        ),  # Ballot Box With Ballot
        (
            0x1F5FA,
            0x1F5FA,
        ),  # World Map
        (
            0x1F6CB,
            0x1F6CB,
        ),  # Couch And Lamp
        (
            0x1F6CD,
            0x1F6CF,
        ),  # Shopping Bags           ..Bed
        (
            0x1F6E0,
            0x1F6E5,
        ),  # Hammer And Wrench       ..Motor Boat
        (
            0x1F6E9,
            0x1F6E9,
        ),  # Small Airplane
        (
            0x1F6F0,
            0x1F6F0,
        ),  # Satellite
        (
            0x1F6F3,
            0x1F6F3,
        ),  # Passenger Ship
    ),
}
