/**
 * @file alda.h
 * @brief Main header file for the Alda parser library.
 *
 * This header includes all public API headers for convenience.
 */

#ifndef ALDA_H
#define ALDA_H

#include "tokens.h"
#include "error.h"
#include "scanner.h"
#include "ast.h"
#include "parser.h"

#endif /* ALDA_H */
