"""
Exports ZERO_WIDTH table keyed by supporting unicode version level.

This code generated by wcwidth/bin/update-tables.py on 2025-09-18 07:49:05 UTC.
"""

ZERO_WIDTH = {
    "4.1.0": (
        # Source: DerivedGeneralCategory-4.1.0.txt
        # Date: 2005-02-26, 02:35:50 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00486,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Psili
        (
            0x00488,
            0x00489,
        ),  # Combining Cyrillic Hundr..Combining Cyrillic Milli
        (
            0x00591,
            0x005B9,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Holam
        (
            0x005BB,
            0x005BD,
        ),  # Hebrew Point Qubuts     ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00603,
        ),  # Arabic Number Sign      ..Arabic Sign Safha
        (
            0x00610,
            0x00615,
        ),  # Arabic Sign Sallallahou ..Arabic Small High Tah
        (
            0x0064B,
            0x0065E,
        ),  # Arabic Fathatan         ..Arabic Fatha With Two Do
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006E4,
        ),  # Arabic Small High Ligatu..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x00901,
            0x00903,
        ),  # Devanagari Sign Candrabi..Devanagari Sign Visarga
        (
            0x0093C,
            0x0093C,
        ),  # Devanagari Sign Nukta
        (
            0x0093E,
            0x0094D,
        ),  # Devanagari Vowel Sign Aa..Devanagari Sign Virama
        (
            0x00951,
            0x00954,
        ),  # Devanagari Stress Sign U..Devanagari Acute Accent
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B43,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D43,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F90,
            0x00F97,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102C,
            0x01032,
        ),  # Myanmar Vowel Sign Aa   ..Myanmar Vowel Sign Ai
        (
            0x01036,
            0x01039,
        ),  # Myanmar Sign Anusvara   ..Myanmar Sign Virama
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135F,
            0x0135F,
        ),  # Ethiopic Combining Gemination Mark
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01DC0,
            0x01DC3,
        ),  # Combining Dotted Grave A..Combining Suspension Mar
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02063,
        ),  # Word Joiner             ..Invisible Separator
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020EB,
        ),  # Combining Left Harpoon A..Combining Long Double So
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE23,
        ),  # Combining Ligature Left ..Combining Double Tilde R
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "5.0.0": (
        # Source: DerivedGeneralCategory-5.0.0.txt
        # Date: 2006-02-27, 23:41:27 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00486,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Psili
        (
            0x00488,
            0x00489,
        ),  # Combining Cyrillic Hundr..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00603,
        ),  # Arabic Number Sign      ..Arabic Sign Safha
        (
            0x00610,
            0x00615,
        ),  # Arabic Sign Sallallahou ..Arabic Small High Tah
        (
            0x0064B,
            0x0065E,
        ),  # Arabic Fathatan         ..Arabic Fatha With Two Do
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006E4,
        ),  # Arabic Small High Ligatu..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00901,
            0x00903,
        ),  # Devanagari Sign Candrabi..Devanagari Sign Visarga
        (
            0x0093C,
            0x0093C,
        ),  # Devanagari Sign Nukta
        (
            0x0093E,
            0x0094D,
        ),  # Devanagari Vowel Sign Aa..Devanagari Sign Virama
        (
            0x00951,
            0x00954,
        ),  # Devanagari Stress Sign U..Devanagari Acute Accent
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B43,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D43,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F90,
            0x00F97,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102C,
            0x01032,
        ),  # Myanmar Vowel Sign Aa   ..Myanmar Vowel Sign Ai
        (
            0x01036,
            0x01039,
        ),  # Myanmar Sign Anusvara   ..Myanmar Sign Virama
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135F,
            0x0135F,
        ),  # Ethiopic Combining Gemination Mark
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01DC0,
            0x01DCA,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFE,
            0x01DFF,
        ),  # Combining Left Arrowhead..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02063,
        ),  # Word Joiner             ..Invisible Separator
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020EF,
        ),  # Combining Left Harpoon A..Combining Right Arrow Be
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE23,
        ),  # Combining Ligature Left ..Combining Double Tilde R
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "5.1.0": (
        # Source: DerivedGeneralCategory-5.1.0.txt
        # Date: 2008-03-20, 17:54:57 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00603,
        ),  # Arabic Number Sign      ..Arabic Sign Safha
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0064B,
            0x0065E,
        ),  # Arabic Fathatan         ..Arabic Fatha With Two Do
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006E4,
        ),  # Arabic Small High Ligatu..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00901,
            0x00903,
        ),  # Devanagari Sign Candrabi..Devanagari Sign Visarga
        (
            0x0093C,
            0x0093C,
        ),  # Devanagari Sign Nukta
        (
            0x0093E,
            0x0094D,
        ),  # Devanagari Vowel Sign Aa..Devanagari Sign Virama
        (
            0x00951,
            0x00954,
        ),  # Devanagari Stress Sign U..Devanagari Acute Accent
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F90,
            0x00F97,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135F,
            0x0135F,
        ),  # Ethiopic Combining Gemination Mark
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAA,
        ),  # Sundanese Consonant Sign..Sundanese Sign Pamaaeh
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFE,
            0x01DFF,
        ),  # Combining Left Arrowhead..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A67C,
            0x0A67D,
        ),  # Combining Cyrillic Kavyk..Combining Cyrillic Payer
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "5.2.0": (
        # Source: DerivedGeneralCategory-5.2.0.txt
        # Date: 2009-08-22, 04:58:21 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00603,
        ),  # Arabic Number Sign      ..Arabic Sign Safha
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0064B,
            0x0065E,
        ),  # Arabic Fathatan         ..Arabic Fatha With Two Do
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006E4,
        ),  # Arabic Small High Ligatu..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00900,
            0x00903,
        ),  # Devanagari Sign Inverted..Devanagari Sign Visarga
        (
            0x0093C,
            0x0093C,
        ),  # Devanagari Sign Nukta
        (
            0x0093E,
            0x0094E,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Pr
        (
            0x00951,
            0x00955,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Ca
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F90,
            0x00F97,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135F,
            0x0135F,
        ),  # Ethiopic Combining Gemination Mark
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAA,
        ),  # Sundanese Consonant Sign..Sundanese Sign Pamaaeh
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF2,
        ),  # Vedic Sign Ardhavisarga
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFD,
            0x01DFF,
        ),  # Combining Almost Equal T..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A67C,
            0x0A67D,
        ),  # Combining Cyrillic Kavyk..Combining Cyrillic Payer
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7B,
        ),  # Myanmar Sign Pao Karen Tone
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x11080,
            0x11082,
        ),  # Kaithi Sign Candrabindu ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "6.0.0": (
        # Source: DerivedGeneralCategory-6.0.0.txt
        # Date: 2010-08-19, 00:48:09 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00603,
        ),  # Arabic Number Sign      ..Arabic Sign Safha
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00900,
            0x00903,
        ),  # Devanagari Sign Inverted..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAA,
        ),  # Sundanese Consonant Sign..Sundanese Sign Pamaaeh
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF2,
        ),  # Vedic Sign Ardhavisarga
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A67C,
            0x0A67D,
        ),  # Combining Cyrillic Kavyk..Combining Cyrillic Payer
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7B,
        ),  # Myanmar Sign Pao Karen Tone
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11080,
            0x11082,
        ),  # Kaithi Sign Candrabindu ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "6.1.0": (
        # Source: DerivedGeneralCategory-6.1.0.txt
        # Date: 2011-11-27, 05:10:22 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00604,
        ),  # Arabic Number Sign      ..Arabic Sign Samvat
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008E4,
            0x008FE,
        ),  # Arabic Curly Fatha      ..Arabic Damma With Dot
        (
            0x00900,
            0x00903,
        ),  # Devanagari Sign Inverted..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69F,
            0x0A69F,
        ),  # Combining Cyrillic Letter Iotified E
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7B,
        ),  # Myanmar Sign Pao Karen Tone
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11080,
            0x11082,
        ),  # Kaithi Sign Candrabindu ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "6.2.0": (
        # Source: DerivedGeneralCategory-6.2.0.txt
        # Date: 2012-05-20, 00:42:34 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00604,
        ),  # Arabic Number Sign      ..Arabic Sign Samvat
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008E4,
            0x008FE,
        ),  # Arabic Curly Fatha      ..Arabic Damma With Dot
        (
            0x00900,
            0x00903,
        ),  # Devanagari Sign Inverted..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180D,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x0206A,
            0x0206F,
        ),  # Inhibit Symmetric Swappi..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69F,
            0x0A69F,
        ),  # Combining Cyrillic Letter Iotified E
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7B,
        ),  # Myanmar Sign Pao Karen Tone
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11080,
            0x11082,
        ),  # Kaithi Sign Candrabindu ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "6.3.0": (
        # Source: DerivedGeneralCategory-6.3.0.txt
        # Date: 2013-07-05, 14:08:45 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00604,
        ),  # Arabic Number Sign      ..Arabic Sign Samvat
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008E4,
            0x008FE,
        ),  # Arabic Curly Fatha      ..Arabic Damma With Dot
        (
            0x00900,
            0x00903,
        ),  # Devanagari Sign Inverted..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C01,
            0x00C03,
        ),  # Telugu Sign Candrabindu ..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C82,
            0x00C83,
        ),  # Kannada Sign Anusvara   ..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D02,
            0x00D03,
        ),  # Malayalam Sign Anusvara ..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01DC0,
            0x01DE6,
        ),  # Combining Dotted Grave A..Combining Latin Small Le
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69F,
            0x0A69F,
        ),  # Combining Cyrillic Letter Iotified E
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7B,
        ),  # Myanmar Sign Pao Karen Tone
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE26,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11080,
            0x11082,
        ),  # Kaithi Sign Candrabindu ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "7.0.0": (
        # Source: DerivedGeneralCategory-7.0.0.txt
        # Date: 2014-02-07, 18:42:12 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008E4,
            0x00903,
        ),  # Arabic Curly Fatha      ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C03,
        ),  # Telugu Sign Combining Ca..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D01,
            0x00D03,
        ),  # Malayalam Sign Candrabin..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x019B0,
            0x019C0,
        ),  # New Tai Lue Vowel Sign V..New Tai Lue Vowel Sign I
        (
            0x019C8,
            0x019C9,
        ),  # New Tai Lue Tone Mark-1 ..New Tai Lue Tone Mark-2
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01CF8,
            0x01CF9,
        ),  # Vedic Tone Ring Above   ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF5,
        ),  # Combining Dotted Grave A..Combining Up Tack Above
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69F,
            0x0A69F,
        ),  # Combining Cyrillic Letter Iotified E
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2D,
        ),  # Combining Ligature Left ..Combining Conjoining Mac
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11301,
            0x11303,
        ),  # Grantha Sign Candrabindu..Grantha Sign Visarga
        (
            0x1133C,
            0x1133C,
        ),  # Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "8.0.0": (
        # Source: DerivedGeneralCategory-8.0.0.txt
        # Date: 2015-02-13, 13:47:11 GMT [MD]
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008E3,
            0x00903,
        ),  # Arabic Turned Damma Belo..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C03,
        ),  # Telugu Sign Combining Ca..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D01,
            0x00D03,
        ),  # Malayalam Sign Candrabin..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01CF8,
            0x01CF9,
        ),  # Vedic Tone Ring Above   ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF5,
        ),  # Combining Dotted Grave A..Combining Up Tack Above
        (
            0x01DFC,
            0x01DFF,
        ),  # Combining Double Inverte..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C4,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Virama
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111CA,
            0x111CC,
        ),  # Sharada Sign Nukta      ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133C,
            0x1133C,
        ),  # Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "9.0.0": (
        # Source: DerivedGeneralCategory-9.0.0.txt
        # Date: 2016-06-01, 10:34:26 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D4,
            0x00903,
        ),  # Arabic Small High Word A..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C03,
        ),  # Telugu Sign Combining Ca..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D01,
            0x00D03,
        ),  # Malayalam Sign Candrabin..Malayalam Sign Visarga
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01CF8,
            0x01CF9,
        ),  # Vedic Tone Ring Above   ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF5,
        ),  # Combining Dotted Grave A..Combining Up Tack Above
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111CA,
            0x111CC,
        ),  # Sharada Sign Nukta      ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133C,
            0x1133C,
        ),  # Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "10.0.0": (
        # Source: DerivedGeneralCategory-10.0.0.txt
        # Date: 2017-03-08, 08:41:49 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D4,
            0x00903,
        ),  # Arabic Small High Word A..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C03,
        ),  # Telugu Sign Combining Ca..Telugu Sign Visarga
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF9,
        ),  # Combining Dotted Grave A..Combining Wide Inverted
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111CA,
            0x111CC,
        ),  # Sharada Sign Nukta      ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133C,
            0x1133C,
        ),  # Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "11.0.0": (
        # Source: DerivedGeneralCategory-11.0.0.txt
        # Date: 2018-02-21, 05:34:04 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D3,
            0x00903,
        ),  # Arabic Small Low Waw    ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EB9,
        ),  # Lao Vowel Sign I        ..Lao Vowel Sign Uu
        (
            0x00EBB,
            0x00EBC,
        ),  # Lao Vowel Sign Mai Kon  ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF2,
            0x01CF4,
        ),  # Vedic Sign Ardhavisarga ..Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF9,
        ),  # Combining Dotted Grave A..Combining Wide Inverted
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F51,
            0x16F7E,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ng
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "12.0.0": (
        # Source: DerivedGeneralCategory-12.0.0.txt
        # Date: 2019-01-22, 08:18:28 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D3,
            0x00903,
        ),  # Arabic Small Low Waw    ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF9,
        ),  # Combining Dotted Grave A..Combining Wide Inverted
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x13430,
            0x13438,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph End
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "12.1.0": (
        # Source: DerivedGeneralCategory-12.1.0.txt
        # Date: 2019-03-10, 10:53:08 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D3,
            0x00903,
        ),  # Arabic Small Low Waw    ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B56,
            0x00B57,
        ),  # Oriya Ai Length Mark    ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D82,
            0x00D83,
        ),  # Sinhala Sign Anusvaraya ..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ABE,
        ),  # Combining Doubled Circum..Combining Parentheses Ov
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF9,
        ),  # Combining Dotted Grave A..Combining Wide Inverted
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x13430,
            0x13438,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph End
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "13.0.0": (
        # Source: DerivedGeneralCategory-13.0.0.txt
        # Date: 2019-10-21, 14:30:32 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x008D3,
            0x00903,
        ),  # Arabic Small Low Waw    ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01714,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Virama
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180E,
        ),  # Mongolian Free Variation..Mongolian Vowel Separato
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01AC0,
        ),  # Combining Doubled Circum..Combining Latin Small Le
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DF9,
        ),  # Combining Dotted Grave A..Combining Wide Inverted
        (
            0x01DFB,
            0x01DFF,
        ),  # Combining Deletion Mark ..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x13430,
            0x13438,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph End
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "14.0.0": (
        # Source: DerivedGeneralCategory-14.0.0.txt
        # Date: 2021-07-10, 00:35:08 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00890,
            0x00891,
        ),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (
            0x00898,
            0x0089F,
        ),  # Arabic Small High Word A..Arabic Half Madda Over M
        (
            0x008CA,
            0x00903,
        ),  # Arabic Small High Farsi ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3C,
            0x00C3C,
        ),  # Telugu Sign Nukta
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECD,
        ),  # Lao Tone Mai Ek         ..Lao Niggahita
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01715,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Pamudpod
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180F,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ACE,
        ),  # Combining Doubled Circum..Combining Latin Small Le
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DFF,
        ),  # Combining Dotted Grave A..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x10F82,
            0x10F85,
        ),  # Old Uyghur Combining Dot..Old Uyghur Combining Two
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11070,
            0x11070,
        ),  # Brahmi Sign Old Tamil Virama
        (
            0x11073,
            0x11074,
        ),  # Brahmi Vowel Sign Old Ta..Brahmi Vowel Sign Old Ta
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110C2,
            0x110C2,
        ),  # Kaithi Vowel Sign Vocalic R
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x13430,
            0x13438,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph End
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1CF00,
            0x1CF2D,
        ),  # Znamenny Combining Mark ..Znamenny Combining Mark
        (
            0x1CF30,
            0x1CF46,
        ),  # Znamenny Combining Tonal..Znamenny Priznak Modifie
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2AE,
            0x1E2AE,
        ),  # Toto Sign Rising Tone
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "15.0.0": (
        # Source: DerivedGeneralCategory-15.0.0.txt
        # Date: 2022-04-26, 23:14:35 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00890,
            0x00891,
        ),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (
            0x00898,
            0x0089F,
        ),  # Arabic Small High Word A..Arabic Half Madda Over M
        (
            0x008CA,
            0x00903,
        ),  # Arabic Small High Farsi ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3C,
            0x00C3C,
        ),  # Telugu Sign Nukta
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00CF3,
            0x00CF3,
        ),  # Kannada Sign Combining Anusvara Above Right
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECE,
        ),  # Lao Tone Mai Ek         ..Lao Yamakkan
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01715,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Pamudpod
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180F,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ACE,
        ),  # Combining Doubled Circum..Combining Latin Small Le
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DFF,
        ),  # Combining Dotted Grave A..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10EFD,
            0x10EFF,
        ),  # Arabic Small Low Word Sa..Arabic Small Low Word Ma
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x10F82,
            0x10F85,
        ),  # Old Uyghur Combining Dot..Old Uyghur Combining Two
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11070,
            0x11070,
        ),  # Brahmi Sign Old Tamil Virama
        (
            0x11073,
            0x11074,
        ),  # Brahmi Vowel Sign Old Ta..Brahmi Vowel Sign Old Ta
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110C2,
            0x110C2,
        ),  # Kaithi Vowel Sign Vocalic R
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x11241,
            0x11241,
        ),  # Khojki Vowel Sign Vocalic R
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x11F00,
            0x11F01,
        ),  # Kawi Sign Candrabindu   ..Kawi Sign Anusvara
        (
            0x11F03,
            0x11F03,
        ),  # Kawi Sign Visarga
        (
            0x11F34,
            0x11F3A,
        ),  # Kawi Vowel Sign Aa      ..Kawi Vowel Sign Vocalic
        (
            0x11F3E,
            0x11F42,
        ),  # Kawi Vowel Sign E       ..Kawi Conjoiner
        (
            0x13430,
            0x13440,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph Mirr
        (
            0x13447,
            0x13455,
        ),  # Egyptian Hieroglyph Modi..Egyptian Hieroglyph Modi
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1CF00,
            0x1CF2D,
        ),  # Znamenny Combining Mark ..Znamenny Combining Mark
        (
            0x1CF30,
            0x1CF46,
        ),  # Znamenny Combining Tonal..Znamenny Priznak Modifie
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E08F,
            0x1E08F,
        ),  # Combining Cyrillic Small Letter Byelorussian-ukr
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2AE,
            0x1E2AE,
        ),  # Toto Sign Rising Tone
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E4EC,
            0x1E4EF,
        ),  # Nag Mundari Sign Muhor  ..Nag Mundari Sign Sutuh
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "15.1.0": (
        # Source: DerivedGeneralCategory-15.1.0.txt
        # Date: 2023-07-28, 23:34:02 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00890,
            0x00891,
        ),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (
            0x00898,
            0x0089F,
        ),  # Arabic Small High Word A..Arabic Half Madda Over M
        (
            0x008CA,
            0x00903,
        ),  # Arabic Small High Farsi ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3C,
            0x00C3C,
        ),  # Telugu Sign Nukta
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00CF3,
            0x00CF3,
        ),  # Kannada Sign Combining Anusvara Above Right
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECE,
        ),  # Lao Tone Mai Ek         ..Lao Yamakkan
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01715,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Pamudpod
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180F,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ACE,
        ),  # Combining Doubled Circum..Combining Latin Small Le
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DFF,
        ),  # Combining Dotted Grave A..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10EFD,
            0x10EFF,
        ),  # Arabic Small Low Word Sa..Arabic Small Low Word Ma
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x10F82,
            0x10F85,
        ),  # Old Uyghur Combining Dot..Old Uyghur Combining Two
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11070,
            0x11070,
        ),  # Brahmi Sign Old Tamil Virama
        (
            0x11073,
            0x11074,
        ),  # Brahmi Vowel Sign Old Ta..Brahmi Vowel Sign Old Ta
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110C2,
            0x110C2,
        ),  # Kaithi Vowel Sign Vocalic R
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x11241,
            0x11241,
        ),  # Khojki Vowel Sign Vocalic R
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x11F00,
            0x11F01,
        ),  # Kawi Sign Candrabindu   ..Kawi Sign Anusvara
        (
            0x11F03,
            0x11F03,
        ),  # Kawi Sign Visarga
        (
            0x11F34,
            0x11F3A,
        ),  # Kawi Vowel Sign Aa      ..Kawi Vowel Sign Vocalic
        (
            0x11F3E,
            0x11F42,
        ),  # Kawi Vowel Sign E       ..Kawi Conjoiner
        (
            0x13430,
            0x13440,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph Mirr
        (
            0x13447,
            0x13455,
        ),  # Egyptian Hieroglyph Modi..Egyptian Hieroglyph Modi
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1CF00,
            0x1CF2D,
        ),  # Znamenny Combining Mark ..Znamenny Combining Mark
        (
            0x1CF30,
            0x1CF46,
        ),  # Znamenny Combining Tonal..Znamenny Priznak Modifie
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E08F,
            0x1E08F,
        ),  # Combining Cyrillic Small Letter Byelorussian-ukr
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2AE,
            0x1E2AE,
        ),  # Toto Sign Rising Tone
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E4EC,
            0x1E4EF,
        ),  # Nag Mundari Sign Muhor  ..Nag Mundari Sign Sutuh
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "16.0.0": (
        # Source: DerivedGeneralCategory-16.0.0.txt
        # Date: 2024-04-30, 21:48:17 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00890,
            0x00891,
        ),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (
            0x00897,
            0x0089F,
        ),  # (nil)                   ..Arabic Half Madda Over M
        (
            0x008CA,
            0x00903,
        ),  # Arabic Small High Farsi ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3C,
            0x00C3C,
        ),  # Telugu Sign Nukta
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00CF3,
            0x00CF3,
        ),  # Kannada Sign Combining Anusvara Above Right
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECE,
        ),  # Lao Tone Mai Ek         ..Lao Yamakkan
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01715,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Pamudpod
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180F,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ACE,
        ),  # Combining Doubled Circum..Combining Latin Small Le
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DFF,
        ),  # Combining Dotted Grave A..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10D69,
            0x10D6D,
        ),  # (nil)
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10EFC,
            0x10EFF,
        ),  # (nil)                   ..Arabic Small Low Word Ma
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x10F82,
            0x10F85,
        ),  # Old Uyghur Combining Dot..Old Uyghur Combining Two
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11070,
            0x11070,
        ),  # Brahmi Sign Old Tamil Virama
        (
            0x11073,
            0x11074,
        ),  # Brahmi Vowel Sign Old Ta..Brahmi Vowel Sign Old Ta
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110C2,
            0x110C2,
        ),  # Kaithi Vowel Sign Vocalic R
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x11241,
            0x11241,
        ),  # Khojki Vowel Sign Vocalic R
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x113B8,
            0x113C0,
        ),  # (nil)
        (
            0x113C2,
            0x113C2,
        ),  # (nil)
        (
            0x113C5,
            0x113C5,
        ),  # (nil)
        (
            0x113C7,
            0x113CA,
        ),  # (nil)
        (
            0x113CC,
            0x113D0,
        ),  # (nil)
        (
            0x113D2,
            0x113D2,
        ),  # (nil)
        (
            0x113E1,
            0x113E2,
        ),  # (nil)
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x11F00,
            0x11F01,
        ),  # Kawi Sign Candrabindu   ..Kawi Sign Anusvara
        (
            0x11F03,
            0x11F03,
        ),  # Kawi Sign Visarga
        (
            0x11F34,
            0x11F3A,
        ),  # Kawi Vowel Sign Aa      ..Kawi Vowel Sign Vocalic
        (
            0x11F3E,
            0x11F42,
        ),  # Kawi Vowel Sign E       ..Kawi Conjoiner
        (
            0x11F5A,
            0x11F5A,
        ),  # (nil)
        (
            0x13430,
            0x13440,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph Mirr
        (
            0x13447,
            0x13455,
        ),  # Egyptian Hieroglyph Modi..Egyptian Hieroglyph Modi
        (
            0x1611E,
            0x1612F,
        ),  # (nil)
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1CF00,
            0x1CF2D,
        ),  # Znamenny Combining Mark ..Znamenny Combining Mark
        (
            0x1CF30,
            0x1CF46,
        ),  # Znamenny Combining Tonal..Znamenny Priznak Modifie
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E08F,
            0x1E08F,
        ),  # Combining Cyrillic Small Letter Byelorussian-ukr
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2AE,
            0x1E2AE,
        ),  # Toto Sign Rising Tone
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E4EC,
            0x1E4EF,
        ),  # Nag Mundari Sign Muhor  ..Nag Mundari Sign Sutuh
        (
            0x1E5EE,
            0x1E5EF,
        ),  # (nil)
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
    "17.0.0": (
        # Source: DerivedGeneralCategory-17.0.0.txt
        # Date: 2025-07-24, 00:12:50 GMT
        #
        (
            0x00000,
            0x00000,
        ),  # (nil)
        (
            0x00300,
            0x0036F,
        ),  # Combining Grave Accent  ..Combining Latin Small Le
        (
            0x00483,
            0x00489,
        ),  # Combining Cyrillic Titlo..Combining Cyrillic Milli
        (
            0x00591,
            0x005BD,
        ),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
        (
            0x005BF,
            0x005BF,
        ),  # Hebrew Point Rafe
        (
            0x005C1,
            0x005C2,
        ),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
        (
            0x005C4,
            0x005C5,
        ),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
        (
            0x005C7,
            0x005C7,
        ),  # Hebrew Point Qamats Qatan
        (
            0x00600,
            0x00605,
        ),  # Arabic Number Sign      ..Arabic Number Mark Above
        (
            0x00610,
            0x0061A,
        ),  # Arabic Sign Sallallahou ..Arabic Small Kasra
        (
            0x0061C,
            0x0061C,
        ),  # Arabic Letter Mark
        (
            0x0064B,
            0x0065F,
        ),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
        (
            0x00670,
            0x00670,
        ),  # Arabic Letter Superscript Alef
        (
            0x006D6,
            0x006DD,
        ),  # Arabic Small High Ligatu..Arabic End Of Ayah
        (
            0x006DF,
            0x006E4,
        ),  # Arabic Small High Rounde..Arabic Small High Madda
        (
            0x006E7,
            0x006E8,
        ),  # Arabic Small High Yeh   ..Arabic Small High Noon
        (
            0x006EA,
            0x006ED,
        ),  # Arabic Empty Centre Low ..Arabic Small Low Meem
        (
            0x0070F,
            0x0070F,
        ),  # Syriac Abbreviation Mark
        (
            0x00711,
            0x00711,
        ),  # Syriac Letter Superscript Alaph
        (
            0x00730,
            0x0074A,
        ),  # Syriac Pthaha Above     ..Syriac Barrekh
        (
            0x007A6,
            0x007B0,
        ),  # Thaana Abafili          ..Thaana Sukun
        (
            0x007EB,
            0x007F3,
        ),  # Nko Combining Short High..Nko Combining Double Dot
        (
            0x007FD,
            0x007FD,
        ),  # Nko Dantayalan
        (
            0x00816,
            0x00819,
        ),  # Samaritan Mark In       ..Samaritan Mark Dagesh
        (
            0x0081B,
            0x00823,
        ),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
        (
            0x00825,
            0x00827,
        ),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
        (
            0x00829,
            0x0082D,
        ),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
        (
            0x00859,
            0x0085B,
        ),  # Mandaic Affrication Mark..Mandaic Gemination Mark
        (
            0x00890,
            0x00891,
        ),  # Arabic Pound Mark Above ..Arabic Piastre Mark Abov
        (
            0x00897,
            0x0089F,
        ),  # (nil)                   ..Arabic Half Madda Over M
        (
            0x008CA,
            0x00903,
        ),  # Arabic Small High Farsi ..Devanagari Sign Visarga
        (
            0x0093A,
            0x0093C,
        ),  # Devanagari Vowel Sign Oe..Devanagari Sign Nukta
        (
            0x0093E,
            0x0094F,
        ),  # Devanagari Vowel Sign Aa..Devanagari Vowel Sign Aw
        (
            0x00951,
            0x00957,
        ),  # Devanagari Stress Sign U..Devanagari Vowel Sign Uu
        (
            0x00962,
            0x00963,
        ),  # Devanagari Vowel Sign Vo..Devanagari Vowel Sign Vo
        (
            0x00981,
            0x00983,
        ),  # Bengali Sign Candrabindu..Bengali Sign Visarga
        (
            0x009BC,
            0x009BC,
        ),  # Bengali Sign Nukta
        (
            0x009BE,
            0x009C4,
        ),  # Bengali Vowel Sign Aa   ..Bengali Vowel Sign Vocal
        (
            0x009C7,
            0x009C8,
        ),  # Bengali Vowel Sign E    ..Bengali Vowel Sign Ai
        (
            0x009CB,
            0x009CD,
        ),  # Bengali Vowel Sign O    ..Bengali Sign Virama
        (
            0x009D7,
            0x009D7,
        ),  # Bengali Au Length Mark
        (
            0x009E2,
            0x009E3,
        ),  # Bengali Vowel Sign Vocal..Bengali Vowel Sign Vocal
        (
            0x009FE,
            0x009FE,
        ),  # Bengali Sandhi Mark
        (
            0x00A01,
            0x00A03,
        ),  # Gurmukhi Sign Adak Bindi..Gurmukhi Sign Visarga
        (
            0x00A3C,
            0x00A3C,
        ),  # Gurmukhi Sign Nukta
        (
            0x00A3E,
            0x00A42,
        ),  # Gurmukhi Vowel Sign Aa  ..Gurmukhi Vowel Sign Uu
        (
            0x00A47,
            0x00A48,
        ),  # Gurmukhi Vowel Sign Ee  ..Gurmukhi Vowel Sign Ai
        (
            0x00A4B,
            0x00A4D,
        ),  # Gurmukhi Vowel Sign Oo  ..Gurmukhi Sign Virama
        (
            0x00A51,
            0x00A51,
        ),  # Gurmukhi Sign Udaat
        (
            0x00A70,
            0x00A71,
        ),  # Gurmukhi Tippi          ..Gurmukhi Addak
        (
            0x00A75,
            0x00A75,
        ),  # Gurmukhi Sign Yakash
        (
            0x00A81,
            0x00A83,
        ),  # Gujarati Sign Candrabind..Gujarati Sign Visarga
        (
            0x00ABC,
            0x00ABC,
        ),  # Gujarati Sign Nukta
        (
            0x00ABE,
            0x00AC5,
        ),  # Gujarati Vowel Sign Aa  ..Gujarati Vowel Sign Cand
        (
            0x00AC7,
            0x00AC9,
        ),  # Gujarati Vowel Sign E   ..Gujarati Vowel Sign Cand
        (
            0x00ACB,
            0x00ACD,
        ),  # Gujarati Vowel Sign O   ..Gujarati Sign Virama
        (
            0x00AE2,
            0x00AE3,
        ),  # Gujarati Vowel Sign Voca..Gujarati Vowel Sign Voca
        (
            0x00AFA,
            0x00AFF,
        ),  # Gujarati Sign Sukun     ..Gujarati Sign Two-circle
        (
            0x00B01,
            0x00B03,
        ),  # Oriya Sign Candrabindu  ..Oriya Sign Visarga
        (
            0x00B3C,
            0x00B3C,
        ),  # Oriya Sign Nukta
        (
            0x00B3E,
            0x00B44,
        ),  # Oriya Vowel Sign Aa     ..Oriya Vowel Sign Vocalic
        (
            0x00B47,
            0x00B48,
        ),  # Oriya Vowel Sign E      ..Oriya Vowel Sign Ai
        (
            0x00B4B,
            0x00B4D,
        ),  # Oriya Vowel Sign O      ..Oriya Sign Virama
        (
            0x00B55,
            0x00B57,
        ),  # Oriya Sign Overline     ..Oriya Au Length Mark
        (
            0x00B62,
            0x00B63,
        ),  # Oriya Vowel Sign Vocalic..Oriya Vowel Sign Vocalic
        (
            0x00B82,
            0x00B82,
        ),  # Tamil Sign Anusvara
        (
            0x00BBE,
            0x00BC2,
        ),  # Tamil Vowel Sign Aa     ..Tamil Vowel Sign Uu
        (
            0x00BC6,
            0x00BC8,
        ),  # Tamil Vowel Sign E      ..Tamil Vowel Sign Ai
        (
            0x00BCA,
            0x00BCD,
        ),  # Tamil Vowel Sign O      ..Tamil Sign Virama
        (
            0x00BD7,
            0x00BD7,
        ),  # Tamil Au Length Mark
        (
            0x00C00,
            0x00C04,
        ),  # Telugu Sign Combining Ca..Telugu Sign Combining An
        (
            0x00C3C,
            0x00C3C,
        ),  # Telugu Sign Nukta
        (
            0x00C3E,
            0x00C44,
        ),  # Telugu Vowel Sign Aa    ..Telugu Vowel Sign Vocali
        (
            0x00C46,
            0x00C48,
        ),  # Telugu Vowel Sign E     ..Telugu Vowel Sign Ai
        (
            0x00C4A,
            0x00C4D,
        ),  # Telugu Vowel Sign O     ..Telugu Sign Virama
        (
            0x00C55,
            0x00C56,
        ),  # Telugu Length Mark      ..Telugu Ai Length Mark
        (
            0x00C62,
            0x00C63,
        ),  # Telugu Vowel Sign Vocali..Telugu Vowel Sign Vocali
        (
            0x00C81,
            0x00C83,
        ),  # Kannada Sign Candrabindu..Kannada Sign Visarga
        (
            0x00CBC,
            0x00CBC,
        ),  # Kannada Sign Nukta
        (
            0x00CBE,
            0x00CC4,
        ),  # Kannada Vowel Sign Aa   ..Kannada Vowel Sign Vocal
        (
            0x00CC6,
            0x00CC8,
        ),  # Kannada Vowel Sign E    ..Kannada Vowel Sign Ai
        (
            0x00CCA,
            0x00CCD,
        ),  # Kannada Vowel Sign O    ..Kannada Sign Virama
        (
            0x00CD5,
            0x00CD6,
        ),  # Kannada Length Mark     ..Kannada Ai Length Mark
        (
            0x00CE2,
            0x00CE3,
        ),  # Kannada Vowel Sign Vocal..Kannada Vowel Sign Vocal
        (
            0x00CF3,
            0x00CF3,
        ),  # Kannada Sign Combining Anusvara Above Right
        (
            0x00D00,
            0x00D03,
        ),  # Malayalam Sign Combining..Malayalam Sign Visarga
        (
            0x00D3B,
            0x00D3C,
        ),  # Malayalam Sign Vertical ..Malayalam Sign Circular
        (
            0x00D3E,
            0x00D44,
        ),  # Malayalam Vowel Sign Aa ..Malayalam Vowel Sign Voc
        (
            0x00D46,
            0x00D48,
        ),  # Malayalam Vowel Sign E  ..Malayalam Vowel Sign Ai
        (
            0x00D4A,
            0x00D4D,
        ),  # Malayalam Vowel Sign O  ..Malayalam Sign Virama
        (
            0x00D57,
            0x00D57,
        ),  # Malayalam Au Length Mark
        (
            0x00D62,
            0x00D63,
        ),  # Malayalam Vowel Sign Voc..Malayalam Vowel Sign Voc
        (
            0x00D81,
            0x00D83,
        ),  # Sinhala Sign Candrabindu..Sinhala Sign Visargaya
        (
            0x00DCA,
            0x00DCA,
        ),  # Sinhala Sign Al-lakuna
        (
            0x00DCF,
            0x00DD4,
        ),  # Sinhala Vowel Sign Aela-..Sinhala Vowel Sign Ketti
        (
            0x00DD6,
            0x00DD6,
        ),  # Sinhala Vowel Sign Diga Paa-pilla
        (
            0x00DD8,
            0x00DDF,
        ),  # Sinhala Vowel Sign Gaett..Sinhala Vowel Sign Gayan
        (
            0x00DF2,
            0x00DF3,
        ),  # Sinhala Vowel Sign Diga ..Sinhala Vowel Sign Diga
        (
            0x00E31,
            0x00E31,
        ),  # Thai Character Mai Han-akat
        (
            0x00E34,
            0x00E3A,
        ),  # Thai Character Sara I   ..Thai Character Phinthu
        (
            0x00E47,
            0x00E4E,
        ),  # Thai Character Maitaikhu..Thai Character Yamakkan
        (
            0x00EB1,
            0x00EB1,
        ),  # Lao Vowel Sign Mai Kan
        (
            0x00EB4,
            0x00EBC,
        ),  # Lao Vowel Sign I        ..Lao Semivowel Sign Lo
        (
            0x00EC8,
            0x00ECE,
        ),  # Lao Tone Mai Ek         ..Lao Yamakkan
        (
            0x00F18,
            0x00F19,
        ),  # Tibetan Astrological Sig..Tibetan Astrological Sig
        (
            0x00F35,
            0x00F35,
        ),  # Tibetan Mark Ngas Bzung Nyi Zla
        (
            0x00F37,
            0x00F37,
        ),  # Tibetan Mark Ngas Bzung Sgor Rtags
        (
            0x00F39,
            0x00F39,
        ),  # Tibetan Mark Tsa -phru
        (
            0x00F3E,
            0x00F3F,
        ),  # Tibetan Sign Yar Tshes  ..Tibetan Sign Mar Tshes
        (
            0x00F71,
            0x00F84,
        ),  # Tibetan Vowel Sign Aa   ..Tibetan Mark Halanta
        (
            0x00F86,
            0x00F87,
        ),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
        (
            0x00F8D,
            0x00F97,
        ),  # Tibetan Subjoined Sign L..Tibetan Subjoined Letter
        (
            0x00F99,
            0x00FBC,
        ),  # Tibetan Subjoined Letter..Tibetan Subjoined Letter
        (
            0x00FC6,
            0x00FC6,
        ),  # Tibetan Symbol Padma Gdan
        (
            0x0102B,
            0x0103E,
        ),  # Myanmar Vowel Sign Tall ..Myanmar Consonant Sign M
        (
            0x01056,
            0x01059,
        ),  # Myanmar Vowel Sign Vocal..Myanmar Vowel Sign Vocal
        (
            0x0105E,
            0x01060,
        ),  # Myanmar Consonant Sign M..Myanmar Consonant Sign M
        (
            0x01062,
            0x01064,
        ),  # Myanmar Vowel Sign Sgaw ..Myanmar Tone Mark Sgaw K
        (
            0x01067,
            0x0106D,
        ),  # Myanmar Vowel Sign Weste..Myanmar Sign Western Pwo
        (
            0x01071,
            0x01074,
        ),  # Myanmar Vowel Sign Geba ..Myanmar Vowel Sign Kayah
        (
            0x01082,
            0x0108D,
        ),  # Myanmar Consonant Sign S..Myanmar Sign Shan Counci
        (
            0x0108F,
            0x0108F,
        ),  # Myanmar Sign Rumai Palaung Tone-5
        (
            0x0109A,
            0x0109D,
        ),  # Myanmar Sign Khamti Tone..Myanmar Vowel Sign Aiton
        (
            0x01160,
            0x011FF,
        ),  # Hangul Jungseong Filler ..Hangul Jongseong Ssangni
        (
            0x0135D,
            0x0135F,
        ),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
        (
            0x01712,
            0x01715,
        ),  # Tagalog Vowel Sign I    ..Tagalog Sign Pamudpod
        (
            0x01732,
            0x01734,
        ),  # Hanunoo Vowel Sign I    ..Hanunoo Sign Pamudpod
        (
            0x01752,
            0x01753,
        ),  # Buhid Vowel Sign I      ..Buhid Vowel Sign U
        (
            0x01772,
            0x01773,
        ),  # Tagbanwa Vowel Sign I   ..Tagbanwa Vowel Sign U
        (
            0x017B4,
            0x017D3,
        ),  # Khmer Vowel Inherent Aq ..Khmer Sign Bathamasat
        (
            0x017DD,
            0x017DD,
        ),  # Khmer Sign Atthacan
        (
            0x0180B,
            0x0180F,
        ),  # Mongolian Free Variation..Mongolian Free Variation
        (
            0x01885,
            0x01886,
        ),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
        (
            0x018A9,
            0x018A9,
        ),  # Mongolian Letter Ali Gali Dagalga
        (
            0x01920,
            0x0192B,
        ),  # Limbu Vowel Sign A      ..Limbu Subjoined Letter W
        (
            0x01930,
            0x0193B,
        ),  # Limbu Small Letter Ka   ..Limbu Sign Sa-i
        (
            0x01A17,
            0x01A1B,
        ),  # Buginese Vowel Sign I   ..Buginese Vowel Sign Ae
        (
            0x01A55,
            0x01A5E,
        ),  # Tai Tham Consonant Sign ..Tai Tham Consonant Sign
        (
            0x01A60,
            0x01A7C,
        ),  # Tai Tham Sign Sakot     ..Tai Tham Sign Khuen-lue
        (
            0x01A7F,
            0x01A7F,
        ),  # Tai Tham Combining Cryptogrammic Dot
        (
            0x01AB0,
            0x01ADD,
        ),  # Combining Doubled Circum..(nil)
        (
            0x01AE0,
            0x01AEB,
        ),  # (nil)
        (
            0x01B00,
            0x01B04,
        ),  # Balinese Sign Ulu Ricem ..Balinese Sign Bisah
        (
            0x01B34,
            0x01B44,
        ),  # Balinese Sign Rerekan   ..Balinese Adeg Adeg
        (
            0x01B6B,
            0x01B73,
        ),  # Balinese Musical Symbol ..Balinese Musical Symbol
        (
            0x01B80,
            0x01B82,
        ),  # Sundanese Sign Panyecek ..Sundanese Sign Pangwisad
        (
            0x01BA1,
            0x01BAD,
        ),  # Sundanese Consonant Sign..Sundanese Consonant Sign
        (
            0x01BE6,
            0x01BF3,
        ),  # Batak Sign Tompi        ..Batak Panongonan
        (
            0x01C24,
            0x01C37,
        ),  # Lepcha Subjoined Letter ..Lepcha Sign Nukta
        (
            0x01CD0,
            0x01CD2,
        ),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
        (
            0x01CD4,
            0x01CE8,
        ),  # Vedic Sign Yajurvedic Mi..Vedic Sign Visarga Anuda
        (
            0x01CED,
            0x01CED,
        ),  # Vedic Sign Tiryak
        (
            0x01CF4,
            0x01CF4,
        ),  # Vedic Tone Candra Above
        (
            0x01CF7,
            0x01CF9,
        ),  # Vedic Sign Atikrama     ..Vedic Tone Double Ring A
        (
            0x01DC0,
            0x01DFF,
        ),  # Combining Dotted Grave A..Combining Right Arrowhea
        (
            0x0200B,
            0x0200F,
        ),  # Zero Width Space        ..Right-to-left Mark
        (
            0x02028,
            0x0202E,
        ),  # Line Separator          ..Right-to-left Override
        (
            0x02060,
            0x02064,
        ),  # Word Joiner             ..Invisible Plus
        (
            0x02066,
            0x0206F,
        ),  # Left-to-right Isolate   ..Nominal Digit Shapes
        (
            0x020D0,
            0x020F0,
        ),  # Combining Left Harpoon A..Combining Asterisk Above
        (
            0x02CEF,
            0x02CF1,
        ),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
        (
            0x02D7F,
            0x02D7F,
        ),  # Tifinagh Consonant Joiner
        (
            0x02DE0,
            0x02DFF,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0302A,
            0x0302F,
        ),  # Ideographic Level Tone M..Hangul Double Dot Tone M
        (
            0x03099,
            0x0309A,
        ),  # Combining Katakana-hirag..Combining Katakana-hirag
        (
            0x0A66F,
            0x0A672,
        ),  # Combining Cyrillic Vzmet..Combining Cyrillic Thous
        (
            0x0A674,
            0x0A67D,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Payer
        (
            0x0A69E,
            0x0A69F,
        ),  # Combining Cyrillic Lette..Combining Cyrillic Lette
        (
            0x0A6F0,
            0x0A6F1,
        ),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
        (
            0x0A802,
            0x0A802,
        ),  # Syloti Nagri Sign Dvisvara
        (
            0x0A806,
            0x0A806,
        ),  # Syloti Nagri Sign Hasanta
        (
            0x0A80B,
            0x0A80B,
        ),  # Syloti Nagri Sign Anusvara
        (
            0x0A823,
            0x0A827,
        ),  # Syloti Nagri Vowel Sign ..Syloti Nagri Vowel Sign
        (
            0x0A82C,
            0x0A82C,
        ),  # Syloti Nagri Sign Alternate Hasanta
        (
            0x0A880,
            0x0A881,
        ),  # Saurashtra Sign Anusvara..Saurashtra Sign Visarga
        (
            0x0A8B4,
            0x0A8C5,
        ),  # Saurashtra Consonant Sig..Saurashtra Sign Candrabi
        (
            0x0A8E0,
            0x0A8F1,
        ),  # Combining Devanagari Dig..Combining Devanagari Sig
        (
            0x0A8FF,
            0x0A8FF,
        ),  # Devanagari Vowel Sign Ay
        (
            0x0A926,
            0x0A92D,
        ),  # Kayah Li Vowel Ue       ..Kayah Li Tone Calya Plop
        (
            0x0A947,
            0x0A953,
        ),  # Rejang Vowel Sign I     ..Rejang Virama
        (
            0x0A980,
            0x0A983,
        ),  # Javanese Sign Panyangga ..Javanese Sign Wignyan
        (
            0x0A9B3,
            0x0A9C0,
        ),  # Javanese Sign Cecak Telu..Javanese Pangkon
        (
            0x0A9E5,
            0x0A9E5,
        ),  # Myanmar Sign Shan Saw
        (
            0x0AA29,
            0x0AA36,
        ),  # Cham Vowel Sign Aa      ..Cham Consonant Sign Wa
        (
            0x0AA43,
            0x0AA43,
        ),  # Cham Consonant Sign Final Ng
        (
            0x0AA4C,
            0x0AA4D,
        ),  # Cham Consonant Sign Fina..Cham Consonant Sign Fina
        (
            0x0AA7B,
            0x0AA7D,
        ),  # Myanmar Sign Pao Karen T..Myanmar Sign Tai Laing T
        (
            0x0AAB0,
            0x0AAB0,
        ),  # Tai Viet Mai Kang
        (
            0x0AAB2,
            0x0AAB4,
        ),  # Tai Viet Vowel I        ..Tai Viet Vowel U
        (
            0x0AAB7,
            0x0AAB8,
        ),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
        (
            0x0AABE,
            0x0AABF,
        ),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
        (
            0x0AAC1,
            0x0AAC1,
        ),  # Tai Viet Tone Mai Tho
        (
            0x0AAEB,
            0x0AAEF,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0AAF5,
            0x0AAF6,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Virama
        (
            0x0ABE3,
            0x0ABEA,
        ),  # Meetei Mayek Vowel Sign ..Meetei Mayek Vowel Sign
        (
            0x0ABEC,
            0x0ABED,
        ),  # Meetei Mayek Lum Iyek   ..Meetei Mayek Apun Iyek
        (
            0x0D7B0,
            0x0D7FF,
        ),  # Hangul Jungseong O-yeo  ..(nil)
        (
            0x0FB1E,
            0x0FB1E,
        ),  # Hebrew Point Judeo-spanish Varika
        (
            0x0FE00,
            0x0FE0F,
        ),  # Variation Selector-1    ..Variation Selector-16
        (
            0x0FE20,
            0x0FE2F,
        ),  # Combining Ligature Left ..Combining Cyrillic Titlo
        (
            0x0FEFF,
            0x0FEFF,
        ),  # Zero Width No-break Space
        (
            0x0FFF9,
            0x0FFFB,
        ),  # Interlinear Annotation A..Interlinear Annotation T
        (
            0x101FD,
            0x101FD,
        ),  # Phaistos Disc Sign Combining Oblique Stroke
        (
            0x102E0,
            0x102E0,
        ),  # Coptic Epact Thousands Mark
        (
            0x10376,
            0x1037A,
        ),  # Combining Old Permic Let..Combining Old Permic Let
        (
            0x10A01,
            0x10A03,
        ),  # Kharoshthi Vowel Sign I ..Kharoshthi Vowel Sign Vo
        (
            0x10A05,
            0x10A06,
        ),  # Kharoshthi Vowel Sign E ..Kharoshthi Vowel Sign O
        (
            0x10A0C,
            0x10A0F,
        ),  # Kharoshthi Vowel Length ..Kharoshthi Sign Visarga
        (
            0x10A38,
            0x10A3A,
        ),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
        (
            0x10A3F,
            0x10A3F,
        ),  # Kharoshthi Virama
        (
            0x10AE5,
            0x10AE6,
        ),  # Manichaean Abbreviation ..Manichaean Abbreviation
        (
            0x10D24,
            0x10D27,
        ),  # Hanifi Rohingya Sign Har..Hanifi Rohingya Sign Tas
        (
            0x10D69,
            0x10D6D,
        ),  # (nil)
        (
            0x10EAB,
            0x10EAC,
        ),  # Yezidi Combining Hamza M..Yezidi Combining Madda M
        (
            0x10EFA,
            0x10EFF,
        ),  # (nil)                   ..Arabic Small Low Word Ma
        (
            0x10F46,
            0x10F50,
        ),  # Sogdian Combining Dot Be..Sogdian Combining Stroke
        (
            0x10F82,
            0x10F85,
        ),  # Old Uyghur Combining Dot..Old Uyghur Combining Two
        (
            0x11000,
            0x11002,
        ),  # Brahmi Sign Candrabindu ..Brahmi Sign Visarga
        (
            0x11038,
            0x11046,
        ),  # Brahmi Vowel Sign Aa    ..Brahmi Virama
        (
            0x11070,
            0x11070,
        ),  # Brahmi Sign Old Tamil Virama
        (
            0x11073,
            0x11074,
        ),  # Brahmi Vowel Sign Old Ta..Brahmi Vowel Sign Old Ta
        (
            0x1107F,
            0x11082,
        ),  # Brahmi Number Joiner    ..Kaithi Sign Visarga
        (
            0x110B0,
            0x110BA,
        ),  # Kaithi Vowel Sign Aa    ..Kaithi Sign Nukta
        (
            0x110BD,
            0x110BD,
        ),  # Kaithi Number Sign
        (
            0x110C2,
            0x110C2,
        ),  # Kaithi Vowel Sign Vocalic R
        (
            0x110CD,
            0x110CD,
        ),  # Kaithi Number Sign Above
        (
            0x11100,
            0x11102,
        ),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
        (
            0x11127,
            0x11134,
        ),  # Chakma Vowel Sign A     ..Chakma Maayyaa
        (
            0x11145,
            0x11146,
        ),  # Chakma Vowel Sign Aa    ..Chakma Vowel Sign Ei
        (
            0x11173,
            0x11173,
        ),  # Mahajani Sign Nukta
        (
            0x11180,
            0x11182,
        ),  # Sharada Sign Candrabindu..Sharada Sign Visarga
        (
            0x111B3,
            0x111C0,
        ),  # Sharada Vowel Sign Aa   ..Sharada Sign Virama
        (
            0x111C9,
            0x111CC,
        ),  # Sharada Sandhi Mark     ..Sharada Extra Short Vowe
        (
            0x111CE,
            0x111CF,
        ),  # Sharada Vowel Sign Prish..Sharada Sign Inverted Ca
        (
            0x1122C,
            0x11237,
        ),  # Khojki Vowel Sign Aa    ..Khojki Sign Shadda
        (
            0x1123E,
            0x1123E,
        ),  # Khojki Sign Sukun
        (
            0x11241,
            0x11241,
        ),  # Khojki Vowel Sign Vocalic R
        (
            0x112DF,
            0x112EA,
        ),  # Khudawadi Sign Anusvara ..Khudawadi Sign Virama
        (
            0x11300,
            0x11303,
        ),  # Grantha Sign Combining A..Grantha Sign Visarga
        (
            0x1133B,
            0x1133C,
        ),  # Combining Bindu Below   ..Grantha Sign Nukta
        (
            0x1133E,
            0x11344,
        ),  # Grantha Vowel Sign Aa   ..Grantha Vowel Sign Vocal
        (
            0x11347,
            0x11348,
        ),  # Grantha Vowel Sign Ee   ..Grantha Vowel Sign Ai
        (
            0x1134B,
            0x1134D,
        ),  # Grantha Vowel Sign Oo   ..Grantha Sign Virama
        (
            0x11357,
            0x11357,
        ),  # Grantha Au Length Mark
        (
            0x11362,
            0x11363,
        ),  # Grantha Vowel Sign Vocal..Grantha Vowel Sign Vocal
        (
            0x11366,
            0x1136C,
        ),  # Combining Grantha Digit ..Combining Grantha Digit
        (
            0x11370,
            0x11374,
        ),  # Combining Grantha Letter..Combining Grantha Letter
        (
            0x113B8,
            0x113C0,
        ),  # (nil)
        (
            0x113C2,
            0x113C2,
        ),  # (nil)
        (
            0x113C5,
            0x113C5,
        ),  # (nil)
        (
            0x113C7,
            0x113CA,
        ),  # (nil)
        (
            0x113CC,
            0x113D0,
        ),  # (nil)
        (
            0x113D2,
            0x113D2,
        ),  # (nil)
        (
            0x113E1,
            0x113E2,
        ),  # (nil)
        (
            0x11435,
            0x11446,
        ),  # Newa Vowel Sign Aa      ..Newa Sign Nukta
        (
            0x1145E,
            0x1145E,
        ),  # Newa Sandhi Mark
        (
            0x114B0,
            0x114C3,
        ),  # Tirhuta Vowel Sign Aa   ..Tirhuta Sign Nukta
        (
            0x115AF,
            0x115B5,
        ),  # Siddham Vowel Sign Aa   ..Siddham Vowel Sign Vocal
        (
            0x115B8,
            0x115C0,
        ),  # Siddham Vowel Sign E    ..Siddham Sign Nukta
        (
            0x115DC,
            0x115DD,
        ),  # Siddham Vowel Sign Alter..Siddham Vowel Sign Alter
        (
            0x11630,
            0x11640,
        ),  # Modi Vowel Sign Aa      ..Modi Sign Ardhacandra
        (
            0x116AB,
            0x116B7,
        ),  # Takri Sign Anusvara     ..Takri Sign Nukta
        (
            0x1171D,
            0x1172B,
        ),  # Ahom Consonant Sign Medi..Ahom Sign Killer
        (
            0x1182C,
            0x1183A,
        ),  # Dogra Vowel Sign Aa     ..Dogra Sign Nukta
        (
            0x11930,
            0x11935,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign E
        (
            0x11937,
            0x11938,
        ),  # Dives Akuru Vowel Sign A..Dives Akuru Vowel Sign O
        (
            0x1193B,
            0x1193E,
        ),  # Dives Akuru Sign Anusvar..Dives Akuru Virama
        (
            0x11940,
            0x11940,
        ),  # Dives Akuru Medial Ya
        (
            0x11942,
            0x11943,
        ),  # Dives Akuru Medial Ra   ..Dives Akuru Sign Nukta
        (
            0x119D1,
            0x119D7,
        ),  # Nandinagari Vowel Sign A..Nandinagari Vowel Sign V
        (
            0x119DA,
            0x119E0,
        ),  # Nandinagari Vowel Sign E..Nandinagari Sign Virama
        (
            0x119E4,
            0x119E4,
        ),  # Nandinagari Vowel Sign Prishthamatra E
        (
            0x11A01,
            0x11A0A,
        ),  # Zanabazar Square Vowel S..Zanabazar Square Vowel L
        (
            0x11A33,
            0x11A39,
        ),  # Zanabazar Square Final C..Zanabazar Square Sign Vi
        (
            0x11A3B,
            0x11A3E,
        ),  # Zanabazar Square Cluster..Zanabazar Square Cluster
        (
            0x11A47,
            0x11A47,
        ),  # Zanabazar Square Subjoiner
        (
            0x11A51,
            0x11A5B,
        ),  # Soyombo Vowel Sign I    ..Soyombo Vowel Length Mar
        (
            0x11A8A,
            0x11A99,
        ),  # Soyombo Final Consonant ..Soyombo Subjoiner
        (
            0x11B60,
            0x11B67,
        ),  # (nil)
        (
            0x11C2F,
            0x11C36,
        ),  # Bhaiksuki Vowel Sign Aa ..Bhaiksuki Vowel Sign Voc
        (
            0x11C38,
            0x11C3F,
        ),  # Bhaiksuki Vowel Sign E  ..Bhaiksuki Sign Virama
        (
            0x11C92,
            0x11CA7,
        ),  # Marchen Subjoined Letter..Marchen Subjoined Letter
        (
            0x11CA9,
            0x11CB6,
        ),  # Marchen Subjoined Letter..Marchen Sign Candrabindu
        (
            0x11D31,
            0x11D36,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3A,
            0x11D3A,
        ),  # Masaram Gondi Vowel Sign E
        (
            0x11D3C,
            0x11D3D,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Vowel Sign
        (
            0x11D3F,
            0x11D45,
        ),  # Masaram Gondi Vowel Sign..Masaram Gondi Virama
        (
            0x11D47,
            0x11D47,
        ),  # Masaram Gondi Ra-kara
        (
            0x11D8A,
            0x11D8E,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D90,
            0x11D91,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Vowel Sign
        (
            0x11D93,
            0x11D97,
        ),  # Gunjala Gondi Vowel Sign..Gunjala Gondi Virama
        (
            0x11EF3,
            0x11EF6,
        ),  # Makasar Vowel Sign I    ..Makasar Vowel Sign O
        (
            0x11F00,
            0x11F01,
        ),  # Kawi Sign Candrabindu   ..Kawi Sign Anusvara
        (
            0x11F03,
            0x11F03,
        ),  # Kawi Sign Visarga
        (
            0x11F34,
            0x11F3A,
        ),  # Kawi Vowel Sign Aa      ..Kawi Vowel Sign Vocalic
        (
            0x11F3E,
            0x11F42,
        ),  # Kawi Vowel Sign E       ..Kawi Conjoiner
        (
            0x11F5A,
            0x11F5A,
        ),  # (nil)
        (
            0x13430,
            0x13440,
        ),  # Egyptian Hieroglyph Vert..Egyptian Hieroglyph Mirr
        (
            0x13447,
            0x13455,
        ),  # Egyptian Hieroglyph Modi..Egyptian Hieroglyph Modi
        (
            0x1611E,
            0x1612F,
        ),  # (nil)
        (
            0x16AF0,
            0x16AF4,
        ),  # Bassa Vah Combining High..Bassa Vah Combining High
        (
            0x16B30,
            0x16B36,
        ),  # Pahawh Hmong Mark Cim Tu..Pahawh Hmong Mark Cim Ta
        (
            0x16F4F,
            0x16F4F,
        ),  # Miao Sign Consonant Modifier Bar
        (
            0x16F51,
            0x16F87,
        ),  # Miao Sign Aspiration    ..Miao Vowel Sign Ui
        (
            0x16F8F,
            0x16F92,
        ),  # Miao Tone Right         ..Miao Tone Below
        (
            0x16FE4,
            0x16FE4,
        ),  # Khitan Small Script Filler
        (
            0x16FF0,
            0x16FF1,
        ),  # Vietnamese Alternate Rea..Vietnamese Alternate Rea
        (
            0x1BC9D,
            0x1BC9E,
        ),  # Duployan Thick Letter Se..Duployan Double Mark
        (
            0x1BCA0,
            0x1BCA3,
        ),  # Shorthand Format Letter ..Shorthand Format Up Step
        (
            0x1CF00,
            0x1CF2D,
        ),  # Znamenny Combining Mark ..Znamenny Combining Mark
        (
            0x1CF30,
            0x1CF46,
        ),  # Znamenny Combining Tonal..Znamenny Priznak Modifie
        (
            0x1D165,
            0x1D169,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D16D,
            0x1D182,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D185,
            0x1D18B,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D1AA,
            0x1D1AD,
        ),  # Musical Symbol Combining..Musical Symbol Combining
        (
            0x1D242,
            0x1D244,
        ),  # Combining Greek Musical ..Combining Greek Musical
        (
            0x1DA00,
            0x1DA36,
        ),  # Signwriting Head Rim    ..Signwriting Air Sucking
        (
            0x1DA3B,
            0x1DA6C,
        ),  # Signwriting Mouth Closed..Signwriting Excitement
        (
            0x1DA75,
            0x1DA75,
        ),  # Signwriting Upper Body Tilting From Hip Joints
        (
            0x1DA84,
            0x1DA84,
        ),  # Signwriting Location Head Neck
        (
            0x1DA9B,
            0x1DA9F,
        ),  # Signwriting Fill Modifie..Signwriting Fill Modifie
        (
            0x1DAA1,
            0x1DAAF,
        ),  # Signwriting Rotation Mod..Signwriting Rotation Mod
        (
            0x1E000,
            0x1E006,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E008,
            0x1E018,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E01B,
            0x1E021,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E023,
            0x1E024,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E026,
            0x1E02A,
        ),  # Combining Glagolitic Let..Combining Glagolitic Let
        (
            0x1E08F,
            0x1E08F,
        ),  # Combining Cyrillic Small Letter Byelorussian-ukr
        (
            0x1E130,
            0x1E136,
        ),  # Nyiakeng Puachue Hmong T..Nyiakeng Puachue Hmong T
        (
            0x1E2AE,
            0x1E2AE,
        ),  # Toto Sign Rising Tone
        (
            0x1E2EC,
            0x1E2EF,
        ),  # Wancho Tone Tup         ..Wancho Tone Koini
        (
            0x1E4EC,
            0x1E4EF,
        ),  # Nag Mundari Sign Muhor  ..Nag Mundari Sign Sutuh
        (
            0x1E5EE,
            0x1E5EF,
        ),  # (nil)
        (
            0x1E6E3,
            0x1E6E3,
        ),  # (nil)
        (
            0x1E6E6,
            0x1E6E6,
        ),  # (nil)
        (
            0x1E6EE,
            0x1E6EF,
        ),  # (nil)
        (
            0x1E6F5,
            0x1E6F5,
        ),  # (nil)
        (
            0x1E8D0,
            0x1E8D6,
        ),  # Mende Kikakui Combining ..Mende Kikakui Combining
        (
            0x1E944,
            0x1E94A,
        ),  # Adlam Alif Lengthener   ..Adlam Nukta
        (
            0x1F3FB,
            0x1F3FF,
        ),  # Emoji Modifier Fitzpatri..Emoji Modifier Fitzpatri
        (
            0xE0001,
            0xE0001,
        ),  # Language Tag
        (
            0xE0020,
            0xE007F,
        ),  # Tag Space               ..Cancel Tag
        (
            0xE0100,
            0xE01EF,
        ),  # Variation Selector-17   ..Variation Selector-256
    ),
}
